/*
 * AdjancencySubmatrix implements the maptree functions in the 2020 paper by Worboys and Duckham
 */
package org.duckham.maptrees;

import java.util.*;

/**
 *
 * @author mduckham
 */
public class AdjacencySubmatrix {

    /**
     * Contains the core data for the adjacency submatrix
     */
    protected byte data[][];

    /**
     *
     * Constructor creates empty adjacency submatrix with specified size
     *
     * @param rows number of rows (blocks)
     * @param cols number of columns (VB nodes)
     */
    public AdjacencySubmatrix(int rows, int cols) {
        // Declare data
        data = new byte[rows][cols];
    }

    /**
     *
     * Constructor creates adjacency submatrix with specified size and populates
     * with data
     *
     * @param s String of length rows x cols of 1s and 0s specifying matrix data
     * in row order (e.g. s="111001"for 3 x 2 adjacency submatrix)
     * @param rows number of rows (blocks)
     * @param cols number of columns (VB nodes)
     */
    public AdjacencySubmatrix(String s, int rows, int cols) {
        // Declare data
        data = new byte[rows][cols];

        //  Read in data from String s         
        for (int i = 0; i < rows; i++) {
            setRow(s.substring(i * cols, (i + 1) * cols), i);
        }

        // Convert to canonical form
        toCanonicalForm();
    }

    /**
     *
     * setRow populates a single row with data, overriding any data already
     * stored at that row. Any inconsistent our invalid data is ignored.
     *
     * @param d String of 1s and 0s representing row data
     * @param row index of row to update
     * @return true if data successfully stored, false otherwise
     */
    public boolean setRow(String d, int row) {
        if (row < data.length) {
            if (d.length() == data[0].length) {
                for (int i = 0; i < data[0].length; i++) {
                    int x = Integer.parseInt(d.substring(i, i + 1));
                    data[row][i] = (byte) x;
                }
                return true;
            }
        }
        return false;
    }

    /**
     *
     * setCol populates a single column with data, overriding any data already
     * stored at that column. Any inconsistent our invalid data is ignored.
     *
     * @param d String of 1s and 0s representing column data
     * @param col index of column to update
     * @return true if data successfully stored, false otherwise
     */
    public boolean setCol(String d, int col) {
        if (col < data[0].length) {
            if (d.length() == data.length) {
                for (int i = 0; i < data.length; i++) {
                    int x = Integer.parseInt(d.substring(i, i + 1));
                    data[i][col] = (byte) x;
                }
                return true;
            }
        }
        return false;
    }

    /**
     *
     * extractRow extracts a String of 1s and 0s representing the data stored in
     * a specified row
     *
     * @param row index of row to retrieve data from
     * @return String of 1s and 0s representing row data
     */
    public String extractRow(int row) {
        String result = "";
        if (row < data.length) {
            for (int i = 0; i < data[0].length; i++) {
                result = result + data[row][i];
            }
        }
        return result;
    }

    /**
     *
     * extractCol extracts a String of 1s and 0s representing the data stored in
     * a specified column
     *
     * @param col index of column to retrieve data from
     * @return String of 1s and 0s representing column data
     */
    public String extractCol(int col) {
        String result = "";
        if (col < data[0].length) {
            for (int i = 0; i < data.length; i++) {
                result = result + data[i][col];
            }
        }
        return result;
    }

    /**
     *
     * switchRow exchanges the data in the specified rows
     *
     * @param r1 index of row 1 to swap
     * @param r2 index of row 2 to swap
     */
    public void switchRow(int r1, int r2) {
        if (Math.max(r2, r1) < data.length) {
            String rt = this.extractRow(r1);
            this.setRow(this.extractRow(r2), r1);
            this.setRow(rt, r2);
        }
    }

    /**
     *
     * switchCol exchanges the data in the specified columns
     *
     * @param c1 index of column 1 to swap
     * @param c2 index of column 2 to swap
     */
    public void switchCol(int c1, int c2) {
        if (Math.max(c2, c1) < data[0].length) {
            String ct = this.extractCol(c1);
            this.setCol(this.extractCol(c2), c1);
            this.setCol(ct, c2);
        }
    }

    /**
     *
     * rows returns the number of rows in the adjacency matrix
     *
     * @return integer total number of rows
     */
    public int rows() {
        return data.length;
    }

    /**
     *
     * cols returns the number of columns in the adjacency matrix
     *
     * @return integer total number of columns
     */
    public int cols() {
        return data[0].length;
    }

    /**
     *
     * countDigits returns the total number of 1s in all rows/columns of this
     * adjacency matrix
     *
     * @return number of 1s in this adjacency matrix
     */
    public int countDigits() {
        int res = 0;
        for (int i = 0; i < data.length; i++) {
            for (int j = 0; j < data[0].length; j++) {
                if (data[i][j] > 0) {
                    res = res + 1;
                }
            }

        }
        return res;
    }

    /**
     *
     * countRowDigits returns the total number of 1s in the specified row of
     * this adjacency matrix
     *
     * @param row index of row to retrieve digits from
     * @return number of 1s in the specified row (returns -1 if row index out of
     * bounds)
     */
    public int countRowDigits(int row) {
        int res = -1;
        if (row < data.length) {
            res = 0;
            for (int i = 0; i < data[0].length; i++) {
                if (data[row][i] > 0) {
                    res = res + 1;
                }
            }
        }
        return res;
    }

    /**
     *
     * countColDigits returns the total number of 1s in the specified column of
     * this adjacency matrix
     *
     * @param col index of column to retrieve digits from
     * @return number of 1s in the specified column (returns -1 if column index
     * out of bounds)
     */
    public int countColDigits(int col) {
        int res = -1;
        if (col < data[0].length) {
            res = 0;
            for (int i = 0; i < data.length; i++) {
                if (data[i][col] > 0) {
                    res = res + 1;
                }
            }
        }
        return res;
    }

    /**
     *
     * rowBinary returns the integer value of the specified row when represented
     * as a binary integer (e.g., row ="01101" returns 13)
     *
     * @param row index of row to retrieve integer value of binary number from
     * from
     * @return integer value of the specified row when represented as a binary
     * integer (returns -1 if row index out of bounds)
     */
    public int rowBinary(int row) {
        int res = -1;
        if (row < rows()) {
            res = 0;
            String b = extractRow(row);

            // Initialise d = 1 for first digit
            int d = 1;

            // Traverse row in reverse order, from small end
            for (int j = b.length(); j >= 1; j--) {

                // Add appropriate binary integer for each 1 encountered
                if (b.substring(j - 1, j).equals("1")) {
                    res = res + d;
                }

                // Increment binary integer (1, 2, 4, 8, ...)
                d = d + d;
            }
        }
        return res;
    }

    /**
     *
     * Checks equality of two adjacency matrices. Method assumes both are in
     * canonical form
     *
     * @param o object to compare with
     * @return true if both adjacency submatrix are in canonical form and
     * identical
     */
    public boolean equals(Object o) {
        try {
            AdjacencySubmatrix as = (AdjacencySubmatrix) o;
            if (as.rows() == rows() && as.cols() == cols()) {
                for (int i = 0; i < rows(); i++) {
                    if (!extractRow(i).equals(as.extractRow(i))) {
                        return false;
                    }
                }
                return true;
            }
        } catch (Exception e) {
            // Return false if any exceptions thrown
        }
        return false;
    }

    /**
     *
     * hashCode returns an integer representation of the adjacency matrix
     * suitable for storing in Java Collections and Sets.
     *
     * @return hash code made up of #rows #digits and the int value of the row
     * order representation of matrix as a binary number
     */
    public int hashCode() {
        String res = toCanonicalString();
        if (res.length() > 18) {
            res = res.substring(0, 18);
        }

        return Integer.parseInt(rows() + "" + cols() + "" + Integer.parseInt(res, 2));
    }

    /**
     *
     * toString gives a simple representation of the data in the adjacency
     * submatrix
     *
     * @return String representation of data in the adjacency submatrix
     */
    public String toString() {
        String s = "";
        for (int i = 0; i < data.length; i++) {
            s = s + "|";
            for (int j = 0; j < data[0].length; j++) {
                s = s + data[i][j] + " ";
            }
            s = s.substring(0, s.length() - 1);
            s = s + "|\n";
        }
        return s;
    }

    /**
     * toCanonicalForm converts an adjacency submatrix into a canonical form.
     * More details are in the paper, but adjacency submatrices can have rows
     * and columns arbitrarily reordered. The canonical form has the largest
     * binary numbers first (row reorder) and in the largest possible binary
     * form (column order)
     */
    public void toCanonicalForm() {

        // Order rows by largest number of bits
        orderRows();

        // Order columns by largest number of bits
        orderCols();

        // Reorder rows by largest binary number
        sizeOrder();

        // Pack bits into numbers so as to make the largest binary numbers highest up in the row order
        boolean cont = true;
        while (cont) {
            cont = packBits();
        }

        // Reorder rows by largest binary number
        sizeOrder();


    }

    /**
     *
     * toCanonicalString generates a single binary string representation of the
     * adjacency submatrix using row order
     *
     * @return single binary string representation of the adjacency submatrix in
     * row order
     */
    public String toCanonicalString() {
        String s = "";
        for (int i = 0; i < data.length; i++) {
            for (int j = 0; j < data[0].length; j++) {
                s = s + data[i][j];
            }
        }
        return s;
    }

    // Private method to order rows in canonical order, in descending number of 1s
    private void orderRows() {

        // Cycle through every row with every other row
        for (int i = 0; i < data.length; i++) {

            // Flag if anything changed
            boolean changed = false;

            // Cycle through rows
            for (int j = 0; j < data.length - 1; j++) {
                if (countRowDigits(j) < countRowDigits(j + 1)) {
                    switchRow(j, j + 1);
                    changed = true;
                }
            }

            // Once no change occurred safe to return
            if (!changed) {
                return;
            }
        }
    }

    // Private method to order cols in canonical order, in descending number of 1s
    private void orderCols() {

        // Cycle through every col with every other col
        for (int i = 0; i < data[0].length; i++) {

            // Flag if anything changed
            boolean changed = false;

            // Cycle through cols
            for (int j = 0; j < data[0].length - 1; j++) {
                if (countColDigits(j) < countColDigits(j + 1)) {
                    switchCol(j, j + 1);
                    changed = true;
                }
            }

            // Once no change occurred safe to return
            if (!changed) {
                return;
            }
        }
    }

    private boolean checkTerminalVB(int row) {
        if (countRowDigits(row) < 2) {
            return false;
        }
        int connectcount = 0;
        for (int i = 0; i < rows(); i++) {
            if (i != row) {
                boolean connects = false;
                for (int j = 0; j < cols(); j++) {
                    if (data[0][j] == 1 && data[i][j] == 1) {
                        if (countRowDigits(i) > 1) {
                            connectcount++;
                            if(connectcount==2)return false;
                            j=cols();
                        }
                    }
                }

            }
        }
        return true;
    }

    // Private method to order rows in canonical order, in descending magnitude of binary digits
    private void sizeOrder() {
        
        // Check first row terminal VB
        if(!checkTerminalVB(0)){
            //System.out.println("Triggered with " + toString());
            switchRow(0,1);
            boolean cont = true;
            while(cont)cont = packBits();
        }

        // Cycle through every row with every other row
        for (int i = 0; i < data.length; i++) {

            // Flag if anything changed
            boolean changed = false;

            // Cycle through every row
            for (int j = 0; j < data.length - 1; j++) {

                // Look for pairs of adjacent rows that can be reordered by binary digit size, subject to row digit constraints
                if (rowBinary(j) < rowBinary(j + 1) && countRowDigits(j) <= countRowDigits(j + 1)) {
                    switchRow(j, j + 1);
                    changed = true;
                }
            }

            // Once no change occurred safe to return
            if (!changed) {
                return;
            }
        }
    }

    // Private method to order colums in canonical order, maximising row binary digit value where possible
    // This method needs to be iterated until it return false to safely fine canonical form
    private boolean packBits() {

        // Cycle through every row
        for (int i = 0; i < data.length; i++) {

            // Cycle through every column
            for (int j = 0; j < data[0].length - 1; j++) {

                // Check if there is a digit that might be reordered respecting column digit counts to make a larger binary number
                if (data[i][j] < data[i][j + 1]) {
                    if (countColDigits(j) <= countColDigits(j + 1)) {

                        // Flag for "make the switch?"
                        boolean mts = true;

                        // Cycle through every row above this one
                        for (int k = 0; k < i; k++) {

                            // Check if making the switch will affect a previous row...
                            if (data[k][j] != data[k][j + 1]) {
                                mts = false;
                            }
                        }

                        // Make the switch if all conditions met
                        if (mts) {
                            this.switchCol(j, j + 1);
                            return true;
                        }
                    }
                }

            }
        }
        return false;
    }

    /**
     *
     * Checks adjacency submatrix condition 1: #1s = has rows + columns - 1
     *
     * @return true if this adjacency submatrix satisfies validity criterion 1
     */
    public boolean validSum() {
        return rows() + cols() == countDigits() + 1;
    }

    /**
     *
     * Checks adjacency submatrix condition 2: No rows have zero 1s
     *
     * @return true if this adjacency submatrix satisfies validity criterion 2
     */
    public boolean validRows() {
        for (int i = 0; i < rows(); i++) {
            if (countRowDigits(i) == 0) {
                return false;
            }
        }
        return true;
    }

    /**
     *
     * Checks adjacency submatrix condition 3: No columns have fewer than 2 1s
     *
     * @return true if this adjacency submatrix satisfies validity criterion 3
     */
    public boolean validCols() {
        for (int i = 0; i < cols(); i++) {
            if (countColDigits(i) < 2) {
                return false;
            }
        }
        return true;
    }

    /**
     *
     * Checks adjacency submatrix condition 4: There exists a path between every
     * pair of nodes
     *
     * @return true if this adjacency submatrix satisfies validity criterion 4
     */
    public boolean validPaths() {

        // Declare and initialize row and column "visited" flags 
        boolean[] c = new boolean[cols()];
        for (int i = 0; i < cols(); i++) {
            c[i] = false;
        }
        boolean[] r = new boolean[rows()];
        for (int i = 0; i < rows(); i++) {
            r[i] = false;
        }

        // New priority queue to store nodes to be visited
        PriorityQueue<Integer> q = new PriorityQueue<Integer>();

        // Start by queuing up row 0 to visit
        q.add(new Integer(0));

        // Continue while unvisited rows
        while (!q.isEmpty()) {

            // Dequeue next row
            int x = q.poll().intValue();

            // Flag visited row
            r[x] = true;

            // Cycle through columns
            for (int j = 0; j < cols(); j++) {

                // Check if this cell is connected and this column is unvisited
                if (data[x][j] == 1 && c[j] == false) {

                    // Flag visited column
                    c[j] = true;

                    // Cycle through each row and queue up and connected, unvisited rows
                    for (int i = 0; i < rows(); i++) {
                        if (r[i] == false && data[i][j] == 1) {
                            q.add(new Integer(i));
                        }
                    }
                }
            }
        }

        // Check if all columns visited
        for (int i = 0; i < cols(); i++) {
            if (c[i] == false) {
                return false;
            }
        }
        return true;

    }

    /**
     *
     * Checks if this adjacency submatrix satisfies all four conditions
     *
     * @return
     */
    public boolean isValid() {
        return validSum() && validRows() && validCols() && validPaths();
    }

    // Brute force permute
    private static void permute(String str, int l, int rows, int cols) {
        if (l == rows * cols - 1) {
            AdjacencySubmatrix as = new AdjacencySubmatrix(str, rows, cols);
            if (as.isValid()) {
                boolean b = resultset.add(as);
                if (b) {
                    System.out.println("Found #" + resultset.size());
                }
            }
        } else {
            for (int i = l; i <= rows * cols - 1; i++) {
                //System.out.println("1."+l + "," + i +"," +str);
                str = swap(str, l, i);

                //System.out.println("2."+l + "," + i +"," +str);
                //if (l == i || !sameChar(str, l, i)) {
                if (canPermuteBruteForce(str, l + 1, rows, cols)) {
                    permute(str, l + 1, rows, cols);
                }
                //}
                //System.out.println("3."+l + "," + i +"," +str);

                str = swap(str, l, i);

                //System.out.println("4."+l + "," + i +"," +str);
            }
        }
    }

    public static void enumerateStructuredValid(int rows, int cols) {
        resultset = new HashSet<AdjacencySubmatrix>();
        checked = new HashSet<String>();
        String s = "";
        for (int i = 0; i < rows * cols; i++) {
            if (i <= cols - 1) {
                s = s + "1";
            } else if (i % cols == 0) {
                s = s + "1";
            } else {
                s = s + "0";
            }
        }
        System.out.println(s);
        permuteStructured(s, rows * cols - 1, rows, cols);
        System.out.println(resultset.size());


        System.out.println(resultset.toString());


    }

    private static void permuteStructured(String str, int l, int rows, int cols) {

        //if (canPermuteStructured(str, l - 1, rows, cols))
        //    permuteStructured(str, l - 1, rows, cols);

        for (int i = l; i > 0; i--) {

            //System.out.print("Visiting: marker "+l + " at index " + i +" on " +str + "[" + checked.size()+"]");

            if (!(str.substring(i - 1, i).equals("0") && str.substring(l, l + 1).equals("1"))) {

                str = swap(str, l, i - 1);

                //System.out.println("2."+l + "," + i +"," +str);
                //if (l == i || !sameChar(str, l, i)) {

                if (canPermuteStructured(str, l - 1, rows, cols)) {
                    //System.out.print(" +");
                    //System.out.println();
                    permuteStructured(str, l - 1, rows, cols);
                }
                //}

                str = swap(str, l, i - 1);
            }

            //System.out.println();
        }

        AdjacencySubmatrix as = new AdjacencySubmatrix(str, rows, cols);
        if (as.isValid()) {
            boolean b = resultset.add(as);
            if (b) {
                System.out.println("Found #" + resultset.size() + ": " + str + " ");

            }
        }
    }

    protected static boolean canPermuteStructured(String str, int l, int rows, int cols) {

        boolean matches = false;

        // Store number of 1s in each row
        int rcount[] = new int[rows];
        for (int i = 0; i < rows; i++) {
            rcount[i] = str.substring(i * cols, i * cols + cols).replace("0", "").length();

            // Check for matching adjacent rows
            if (i < cols - 2) {
                if (str.substring(i, i + cols).equals(str.substring(i + cols, i + 2 * cols))) {
                    matches = true;
                }
            }
        }

        // Check at least two 1s in first row 
        if (!str.substring(0, 2).equals("11")) {
            return false;
        }

        // Check at least one 1 left in every row
        for (int i = 0; i < rows; i++) {
            if (rcount[i] < 1) {
                return false;
            }

            // Check for matching 2-adjacent rows
            if (!matches) {
                if (i < cols - 3) {
                    if (str.substring(i, i + cols).equals(str.substring(i + 2 * cols, i + 3 * cols))) {
                        matches = true;
                    }
                }
            }
        }

        // Check every row has at most one more than every subsequent row (strict adherence cuts out some cases)
        for (int i = 0; i < rows - 1; i++) {
            if (rcount[i] + 1 < rcount[i + 1]) {
                //System.out.println("Triggered with: " + str);
                return false;
            }
        }

        // Check top (cols-1) rows are full
        int toprowsum = 0;
        for (int i = 0; i < cols - 1; i++) {
            toprowsum += rcount[i];
        }
        if (toprowsum < (cols - 1) * 2) {
            //System.out.println("Triggered with: " + str);
            return false;
        }



        // Check if already seen
        if (checked.contains(l + "" + str)) {
            return false;
        } else { //if (matches) {
            checked.add(l + "" + str);
        }

        return true;

    }

    // Checks whether a String permutation could lead to valid adjacency submatrix
    private static boolean canPermuteBruteForce(String str, int l, int rows, int cols) {

        // number of rows already passed
        int numrowspassed = (int) Math.floor((double) l / (double) cols);

        // number of 1s already passed
        int sumrowspassed = str.substring(0, numrowspassed * cols).replace("0", "").length();

        // Check if at least one 1 per row passed
        if (sumrowspassed < numrowspassed) {
            return false;
        }

        // Check if at least enough for one 1 per column left to be processed
        if ((rows + cols - 1) - sumrowspassed < rows - numrowspassed) {
            return false;
        }

        return true;
    }

    /**
     *
     * Utility method to swap two characters in a String
     *
     * @param a String to swap characters
     * @param i index of character i in String
     * @param j index of character j in String
     * @return String with characters i and j swapped
     */
    public static String swap(String a, int i, int j) {
        char temp;
        char[] charArray = a.toCharArray();
        temp = charArray[i];
        charArray[i] = charArray[j];
        charArray[j] = temp;
        return String.valueOf(charArray);
    }

    /**
     *
     * Utility method for checking if two indexed characters in String are the
     * same
     *
     * @param a String to check characters
     * @param i index of character i in String
     * @param j index of character j in String
     * @return true if characters i and j in String a are the same
     */
    public static boolean sameChar(String a, int i, int j) {
        char[] charArray = a.toCharArray();
        return charArray[i] == charArray[j];
    }

    public static void enumerateBruteForceValid(int rows, int cols) {
        resultset = new HashSet<AdjacencySubmatrix>();
        String s = "";
        for (int i = 0; i < rows * cols; i++) {
            if (i > rows * cols - (rows + cols)) {
                s = s + "1";
            } else {
                s = s + "0";
            }
        }
        //System.out.println(s);
        permute(s, 0, rows, cols);
        System.out.println(resultset.size());

//        Iterator i = resultset.iterator();
//        HashSet<AdjacencySubmatrix> asresults = new HashSet<AdjacencySubmatrix>();

//        while(i.hasNext()){
//            String r = (String)i.next();
//            //System.out.println(r);
//            AdjacencySubmatrix as = new AdjacencySubmatrix(r,rows,cols);
//            as.toCanonicalForm();
//            if(as.isValid()){
        //System.out.println(asresults.contains(as));
//                boolean added = asresults.add(as);
        //if(added)System.out.println(as.toString());
//            }
//        }
        System.out.println(resultset.toString());


    }
    public static HashSet<AdjacencySubmatrix> resultset;
    public static HashSet<String> checked;
}
